from ophyd import Component as Cpt
from ophyd import EpicsMotor, EpicsSignal


class EpicsMotorEx(EpicsMotor):
    """Extend EpicsMotor with extra configuration fields.
    motor_done_move
    motor_is_moving
    """

    # configuration
    motor_resolution = Cpt(EpicsSignal, ".MRES", kind="config", auto_monitor=True)
    base_velocity = Cpt(EpicsSignal, ".VBAS", kind="config", auto_monitor=True)
    backlash_distance = Cpt(EpicsSignal, ".BDST", kind="config", auto_monitor=True)

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        # get configuration attributes from kwargs and then remove them
        attrs = {}
        for key, value in kwargs.items():
            if hasattr(EpicsMotorEx, key) and isinstance(getattr(EpicsMotorEx, key), Cpt):
                attrs[key] = value
        for key in attrs:
            kwargs.pop(key)

        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )

        # set configuration attributes
        for key, value in attrs.items():
            # print out attributes that are being configured
            print("setting ", key, "=", value)
            getattr(self, key).put(value)

        # self.motor_done_move.subscribe(self._progress_update, run=False)

    # def kickoff(self) -> DeviceStatus:
    #     status = DeviceStatus(self)
    #     self.move(
    #         self._kickoff_params.get("position"),
    #         wait = False
    #     )
    #     return status

    # def _progress_update(self, value, **kwargs) -> None:
    #     self._run_subs(
    #         sub_type=self.SUB_PROGRESS,
    #         value=value ,
    #         done= 1,
    #     )



"""Utility handler to run tasks (function, conditions) in an asynchronous fashion."""

import ctypes
import threading
import traceback
import uuid
from enum import Enum
from typing import TYPE_CHECKING, Callable

from bec_lib.file_utils import get_full_path
from bec_lib.logger import bec_logger
from bec_lib.utils.import_utils import lazy_import_from
from ophyd import Device, DeviceStatus

if TYPE_CHECKING:  # pragma: no cover
    from bec_lib.messages import ScanStatusMessage
else:
    # TODO: put back normal import when Pydantic gets faster
    ScanStatusMessage = lazy_import_from("bec_lib.messages", ("ScanStatusMessage",))


logger = bec_logger.logger

set_async_exc = ctypes.pythonapi.PyThreadState_SetAsyncExc


class TaskState(str, Enum):
    """Possible task states"""

    NOT_STARTED = "not_started"
    RUNNING = "running"
    TIMEOUT = "timeout"
    ERROR = "error"
    COMPLETED = "completed"
    KILLED = "killed"


class TaskKilledError(Exception):
    """Exception raised when a task thread is killed"""


class TaskStatus(DeviceStatus):
    """Thin wrapper around StatusBase to add information about tasks"""

    def __init__(self, device: Device, *, timeout=None, settle_time=0, done=None, success=None):
        super().__init__(
            device=device, timeout=timeout, settle_time=settle_time, done=done, success=success
        )
        self._state = TaskState.NOT_STARTED
        self._task_id = str(uuid.uuid4())

    @property
    def state(self) -> str:
        """Get the state of the task"""
        return self._state.value

    @state.setter
    def state(self, value: TaskState):
        self._state = TaskState(value)

    @property
    def task_id(self) -> str:
        """Get the task ID"""
        return self._task_id


class TaskHandler:
    """Handler to manage asynchronous tasks"""

    def __init__(self, parent: Device):
        """Initialize the handler"""
        self._tasks = {}
        self._parent = parent
        self._lock = threading.RLock()

    def submit_task(
        self,
        task: Callable,
        task_args: tuple | None = None,
        task_kwargs: dict | None = None,
        run: bool = True,
    ) -> TaskStatus:
        """Submit a task to the task handler.

        Args:
            task: The task to run.
            run: Whether to run the task immediately.
        """
        task_args = task_args if task_args else ()
        task_kwargs = task_kwargs if task_kwargs else {}
        task_status = TaskStatus(device=self._parent)
        thread = threading.Thread(
            target=self._wrap_task,
            args=(task, task_args, task_kwargs, task_status),
            name=f"task {task_status.task_id}",
            daemon=True,
        )
        self._tasks.update({task_status.task_id: (task_status, thread)})
        if run is True:
            self.start_task(task_status)
        return task_status

    def start_task(self, task_status: TaskStatus) -> None:
        """Start a task,

        Args:
            task_status: The task status object.
        """
        thread = self._tasks[task_status.task_id][1]
        if thread.is_alive():
            logger.warning(f"Task with ID {task_status.task_id} is already running.")
            return
        task_status.state = TaskState.RUNNING
        thread.start()

    def _wrap_task(
        self, task: Callable, task_args: tuple, task_kwargs: dict, task_status: TaskStatus
    ):
        """Wrap the task in a function"""
        try:
            task(*task_args, **task_kwargs)
        except TimeoutError as exc:
            content = traceback.format_exc()
            logger.warning(
                (
                    f"Timeout Exception in task handler for task {task_status.task_id},"
                    f" Traceback: {content}"
                )
            )
            task_status.state = TaskState.TIMEOUT
            task_status.set_exception(exc)
        except TaskKilledError as exc:
            exc = exc.__class__(
                f"Task {task_status.task_id} was killed. ThreadID:"
                f" {self._tasks[task_status.task_id][1].ident}"
            )
            content = traceback.format_exc()
            logger.warning(
                (
                    f"TaskKilled Exception in task handler for task {task_status.task_id},"
                    f" Traceback: {content}"
                )
            )
            task_status.state = TaskState.KILLED
            task_status.set_exception(exc)
        except Exception as exc:  # pylint: disable=broad-except
            content = traceback.format_exc()
            logger.warning(
                f"Exception in task handler for task {task_status.task_id}, Traceback: {content}"
            )
            task_status.state = TaskState.ERROR
            task_status.set_exception(exc)
        else:
            task_status.state = TaskState.COMPLETED
            task_status.set_finished()
        finally:
            with self._lock:
                self._tasks.pop(task_status.task_id)

    def kill_task(self, task_status: TaskStatus) -> None:
        """Kill the thread

        task_status: The task status object.
        """
        thread = self._tasks[task_status.task_id][1]
        exception_cls = TaskKilledError

        ident = ctypes.c_long(thread.ident)
        exc = ctypes.py_object(exception_cls)
        try:
            res = set_async_exc(ident, exc)
            if res == 0:
                raise ValueError("Invalid thread ID")
            elif res > 1:
                set_async_exc(ident, None)
                logger.warning(f"Exception raise while kille Thread {ident}; return value: {res}")
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f"Exception raised while killing thread {ident}: {e}")

    def shutdown(self):
        """Shutdown all tasks of task handler"""
        with self._lock:
            for info in self._tasks.values():
                self.kill_task(info[0])


class FileHandler:
    """Utility class for file operations."""

    def get_full_path(
        self, scan_status_msg: ScanStatusMessage, name: str, create_dir: bool = True
    ) -> str:
        """Get the file path.

        Args:
            scan_info_msg: The scan info message.
            name: The name of the file.
            create_dir: Whether to create the directory.
        """
        return get_full_path(scan_status_msg=scan_status_msg, name=name, create_dir=create_dir)



"""Utilities to mock and test devices."""

from typing import TYPE_CHECKING
from unittest import mock

from bec_lib.devicemanager import ScanInfo
from bec_lib.logger import bec_logger
from bec_lib.utils.import_utils import lazy_import_from
from ophyd import Device

if TYPE_CHECKING:
    from bec_lib.messages import ScanStatusMessage
else:
    # TODO: put back normal import when Pydantic gets faster
    ScanStatusMessage = lazy_import_from("bec_lib.messages", ("ScanStatusMessage",))

logger = bec_logger.logger


def patch_dual_pvs(device):
    """Patch dual PVs"""
    patch_functions_required_for_connection(device)
    device.wait_for_connection(all_signals=True)
    for walk in device.walk_signals():
        if not hasattr(walk.item, "_read_pv"):
            continue
        if not hasattr(walk.item, "_write_pv"):
            continue
        if walk.item._read_pv.pvname != walk.item._write_pv.pvname:
            walk.item._read_pv = walk.item._write_pv


def patch_functions_required_for_connection(device):
    """Patch functions required for connection. This will run the subs for all sub devices and devices.
    This is needed to ensure that the wait_for_connection method of required for connections methods are properly patched.
    """
    for event in device.event_types:
        device._run_subs(sub_type=event, value=0, timestamp=0)
    for name, dev in device.walk_subdevices(include_lazy=True):
        for event in dev.event_types:
            dev._run_subs(sub_type=event, value=0, timestamp=0)


class SocketMock:
    """Socket Mock. Used for testing"""

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.buffer_put = []
        self.buffer_recv = [b""]
        self.is_open = False
        self.sock = None
        self.open()

    def connect(self):
        """Mock connect method"""
        print(f"connecting to {self.host} port {self.port}")

    def _put(self, msg_bytes):
        """Mock put method"""
        self.buffer_put.append(msg_bytes)
        print(self.buffer_put)

    # pylint: disable=unused-argument
    def _recv(self, buffer_length=1024):
        """Mock receive method"""
        print(self.buffer_recv)
        if isinstance(self.buffer_recv, list):
            if len(self.buffer_recv) > 0:
                ret_val = self.buffer_recv.pop(0)
            else:
                ret_val = b""
            return ret_val
        return self.buffer_recv

    def _initialize_socket(self):
        """Mock initialize socket method"""

    def put(self, msg):
        """Mock put method"""
        return self._put(msg)

    def receive(self, buffer_length=1024):
        """Mock receive method"""
        return self._recv(buffer_length=buffer_length)

    def open(self):
        """Mock open method"""
        self._initialize_socket()
        self.is_open = True

    def close(self):
        """Mock close method"""
        self.sock = None
        self.is_open = False

    def flush_buffer(self):
        """Mock flush buffer method"""
        self.buffer_put = []
        self.buffer_recv = ""


class MockPV:
    """
    MockPV class

    This class is used for mocking pyepics signals for testing purposes

    """

    _fmtsca = "<PV '%(pvname)s', count=%(count)i, type=%(typefull)s, access=%(access)s>"
    _fmtarr = "<PV '%(pvname)s', count=%(count)i/%(nelm)i, type=%(typefull)s, access=%(access)s>"
    _fields = (
        "pvname",
        "value",
        "char_value",
        "status",
        "ftype",
        "chid",
        "host",
        "count",
        "access",
        "write_access",
        "read_access",
        "severity",
        "timestamp",
        "posixseconds",
        "nanoseconds",
        "precision",
        "units",
        "enum_strs",
        "upper_disp_limit",
        "lower_disp_limit",
        "upper_alarm_limit",
        "lower_alarm_limit",
        "lower_warning_limit",
        "upper_warning_limit",
        "upper_ctrl_limit",
        "lower_ctrl_limit",
    )

    def __init__(
        self,
        pvname,
        callback=None,
        form="time",
        verbose=False,
        auto_monitor=None,
        count=None,
        connection_callback=None,
        connection_timeout=None,
        access_callback=None,
    ):
        self.pvname = pvname.strip()
        self.form = form.lower()
        self.verbose = verbose
        self._auto_monitor = auto_monitor
        self.ftype = None
        self.connected = True
        self.connection_timeout = connection_timeout
        self._user_max_count = count

        if self.connection_timeout is None:
            self.connection_timeout = 3
        self._args = {}.fromkeys(self._fields)
        self._args["pvname"] = self.pvname
        self._args["count"] = count
        self._args["nelm"] = -1
        self._args["type"] = "unknown"
        self._args["typefull"] = "unknown"
        self._args["access"] = "unknown"
        self._args["status"] = 0
        self.connection_callbacks = []
        self.mock_data = 0

        if connection_callback is not None:
            self.connection_callbacks = [connection_callback]

        self.access_callbacks = []
        if access_callback is not None:
            self.access_callbacks = [access_callback]

        self.callbacks = {}
        self._put_complete = None
        self._monref = None  # holder of data returned from create_subscription
        self._monref_mask = None
        self._conn_started = False
        if isinstance(callback, (tuple, list)):
            for i, thiscb in enumerate(callback):
                if callable(thiscb):
                    self.callbacks[i] = (thiscb, {})
        elif callable(callback):
            self.callbacks[0] = (callback, {})

        self.chid = None
        self.context = mock.MagicMock()
        self._cache_key = (pvname, form, self.context)
        self._reference_count = 0
        for conn_cb in self.connection_callbacks:
            conn_cb(pvname=pvname, conn=True, pv=self)
        for acc_cb in self.access_callbacks:
            acc_cb(True, True, pv=self)

    # pylint disable: unused-argument
    def wait_for_connection(self, timeout=None):
        """Wait for connection"""
        return self.connected

    # pylint disable: unused-argument
    def get_all_metadata_blocking(self, timeout):
        """Get all metadata blocking"""
        md = self._args.copy()
        md.pop("value", None)
        return md

    def get_all_metadata_callback(self, callback, *, timeout):
        """Get all metadata callback"""

        def get_metadata_thread(pvname):
            md = self.get_all_metadata_blocking(timeout=timeout)
            callback(pvname, md)

        get_metadata_thread(pvname=self.pvname)

    # pylint disable: unused-argument
    def put(
        self, value, wait=False, timeout=None, use_complete=False, callback=None, callback_data=None
    ):
        """MOCK PV, put function"""
        self.mock_data = value
        if callback is not None:
            callback()

    # pylint: disable=unused-argument
    def add_callback(self, callback=None, index=None, run_now=False, with_ctrlvars=True, **kw):
        """Add callback"""
        return mock.MagicMock()

    # pylint: disable=unused-argument
    def get_with_metadata(
        self,
        count=None,
        as_string=False,
        as_numpy=True,
        timeout=None,
        with_ctrlvars=False,
        form=None,
        use_monitor=True,
        as_namespace=False,
    ):
        """Get MOCKPV data together with metadata"""
        return {"value": self.mock_data}

    def get(
        self,
        count=None,
        as_string=False,
        as_numpy=True,
        timeout=None,
        with_ctrlvars=False,
        use_monitor=True,
    ):
        """Get value from MOCKPV"""
        data = self.get_with_metadata(
            count=count,
            as_string=as_string,
            as_numpy=as_numpy,
            timeout=timeout,
            with_ctrlvars=with_ctrlvars,
            use_monitor=use_monitor,
        )
        return data["value"] if data is not None else None


def get_mock_scan_info(device: Device | None) -> ScanInfo:
    """
    Get a mock scan info object.
    """
    return ScanInfo(msg=fake_scan_status_msg(device=device))


def fake_scan_status_msg(device: Device | None = None) -> ScanStatusMessage:
    """
    Create a fake scan status message.

    Args:
        device: The device creating the fake scan status message.

    """
    if device is None:
        device = Device(name="mock_device")
    logger.warning(
        (
            f"Device {device.name} is not connected to a Redis server. Fetching mocked ScanStatusMessage."
        )
    )
    return ScanStatusMessage(
        metadata={},
        scan_id="mock_scan_id",
        status="closed",
        scan_number=0,
        session_id=None,
        num_points=11,
        scan_name="mock_line_scan",
        scan_type="step",
        dataset_number=0,
        scan_report_devices=["samx"],
        user_metadata={},
        readout_priority={
            "monitored": ["bpm4a", "samx"],
            "baseline": ["eyex"],
            "async": ["waveform"],
            "continuous": [],
            "on_request": ["flyer_sim"],
        },
        scan_parameters={
            "exp_time": 0,
            "frames_per_trigger": 1,
            "settling_time": 0,
            "readout_time": 0,
            "optim_trajectory": None,
            "return_to_start": True,
            "relative": True,
            "system_config": {"file_suffix": None, "file_directory": None},
        },
        request_inputs={
            "arg_bundle": ["samx", -10, 10],
            "inputs": {},
            "kwargs": {
                "steps": 11,
                "relative": True,
                "system_config": {"file_suffix": None, "file_directory": None},
            },
        },
        info={
            "readout_priority": {
                "monitored": ["bpm4a", "samx"],
                "baseline": ["eyex"],
                "async": ["waveform"],
                "continuous": [],
                "on_request": ["flyer_sim"],
            },
            "file_suffix": None,
            "file_directory": None,
            "user_metadata": {},
            "RID": "a1d86f61-191c-4460-bcd6-f33c61b395ea",
            "scan_id": "3edb8219-75a7-4791-8f86-d5ca112b771a",
            "queue_id": "0f3639ee-899f-4ad1-9e71-f40514c937ef",
            "scan_motors": ["samx"],
            "num_points": 11,
            "positions": [
                [-10.0],
                [-8.0],
                [-6.0],
                [-4.0],
                [-2.0],
                [0.0],
                [2.0],
                [4.0],
                [6.0],
                [8.0],
                [10.0],
            ],
            "file_path": "./data/test_file",
            "scan_name": "mock_line_scan",
            "scan_type": "step",
            "scan_number": 0,
            "dataset_number": 0,
            "exp_time": 0,
            "frames_per_trigger": 1,
            "settling_time": 0,
            "readout_time": 0,
            "scan_report_devices": ["samx"],
            "monitor_sync": "bec",
            "scan_parameters": {
                "exp_time": 0,
                "frames_per_trigger": 1,
                "settling_time": 0,
                "readout_time": 0,
                "optim_trajectory": None,
                "return_to_start": True,
                "relative": True,
                "system_config": {"file_suffix": None, "file_directory": None},
            },
            "request_inputs": {
                "arg_bundle": ["samx", -10, 10],
                "inputs": {},
                "kwargs": {
                    "steps": 11,
                    "relative": True,
                    "system_config": {"file_suffix": None, "file_directory": None},
                },
            },
            "scan_msgs": [
                "metadata={'file_suffix': None, 'file_directory': None, 'user_metadata': {}, 'RID': 'a1d86f61-191c-4460-bcd6-f33c61b395ea'} scan_type='mock_line_scan' parameter={'args': {'samx': [-10, 10]}, 'kwargs': {'steps': 11, 'relative': True, 'system_config': {'file_suffix': None, 'file_directory': None}}} queue='primary'"
            ],
            "args": {"samx": [-10, 10]},
            "kwargs": {
                "steps": 11,
                "relative": True,
                "system_config": {"file_suffix": None, "file_directory": None},
            },
        },
        timestamp=1737100681.694211,
    )



import functools
import os
from typing import Literal

import requests
from github import Github
from pydantic import BaseModel


class GHConfig(BaseModel):
    token: str
    organization: str
    repository: str
    project_number: int
    graphql_url: str
    rest_url: str
    headers: dict


class ProjectItemHandler:
    """
    A class to handle GitHub project items.
    """

    def __init__(self, gh_config: GHConfig):
        self.gh_config = gh_config
        self.gh = Github(gh_config.token)
        self.repo = self.gh.get_repo(f"{gh_config.organization}/{gh_config.repository}")
        self.project_node_id = self.get_project_node_id()

    def set_issue_status(
        self,
        status: Literal[
            "Selected for Development",
            "Weekly Backlog",
            "In Development",
            "Ready For Review",
            "On Hold",
            "Done",
        ],
        issue_number: int | None = None,
        issue_node_id: str | None = None,
    ):
        """
        Set the status field of a GitHub issue in the project.

        Args:
            status (str): The status to set. Must be one of the predefined statuses.
            issue_number (int, optional): The issue number. If not provided, issue_node_id must be provided.
            issue_node_id (str, optional): The issue node ID. If not provided, issue_number must be provided.
        """
        if not issue_number and not issue_node_id:
            raise ValueError("Either issue_number or issue_node_id must be provided.")
        if issue_number and issue_node_id:
            raise ValueError("Only one of issue_number or issue_node_id must be provided.")
        if issue_number is not None:
            issue = self.repo.get_issue(issue_number)
            issue_id = self.get_issue_info(issue.node_id)[0]["id"]
        else:
            issue_id = issue_node_id
        field_id, option_id = self.get_status_field_id(field_name=status)
        self.set_field_option(issue_id, field_id, option_id)

    def run_graphql(self, query: str, variables: dict) -> dict:
        """
        Execute a GraphQL query against the GitHub API.

        Args:
            query (str): The GraphQL query to execute.
            variables (dict): The variables to pass to the query.

        Returns:
            dict: The response from the GitHub API.
        """
        response = requests.post(
            self.gh_config.graphql_url,
            json={"query": query, "variables": variables},
            headers=self.gh_config.headers,
            timeout=10,
        )
        if response.status_code != 200:
            raise Exception(
                f"Query failed with status code {response.status_code}: {response.text}"
            )
        return response.json()

    def get_project_node_id(self):
        """
        Retrieve the project node ID from the GitHub API.
        """
        query = """
      query($owner: String!, $number: Int!) {
        organization(login: $owner) {
          projectV2(number: $number) {
            id
          }
        }
      }
      """
        variables = {"owner": self.gh_config.organization, "number": self.gh_config.project_number}
        resp = self.run_graphql(query, variables)
        return resp["data"]["organization"]["projectV2"]["id"]

    def get_issue_info(self, issue_node_id: str):
        """
        Get the project-related information for a given issue node ID.

        Args:
            issue_node_id (str): The node ID of the issue. Please note that this is not the issue number and typically starts with "I".

        Returns:
            list[dict]: A list of project items associated with the issue.
        """
        query = """
        query($issueId: ID!) {
          node(id: $issueId) {
            ... on Issue {
              projectItems(first: 10) {
                nodes {
                  project {
                    id
                    title
                  }
                  id
                  fieldValues(first: 20) {
                    nodes {
                      ... on ProjectV2ItemFieldSingleSelectValue {
                        name
                        field {
                          ... on ProjectV2SingleSelectField {
                            name
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
        """
        variables = {"issueId": issue_node_id}
        resp = self.run_graphql(query, variables)
        return resp["data"]["node"]["projectItems"]["nodes"]

    def get_status_field_id(
        self,
        field_name: Literal[
            "Selected for Development",
            "Weekly Backlog",
            "In Development",
            "Ready For Review",
            "On Hold",
            "Done",
        ],
    ) -> tuple[str, str]:
        """
        Get the status field ID and option ID for the given field name in the project.

        Args:
            field_name (str): The name of the field to retrieve.
                Must be one of the predefined statuses.

        Returns:
            tuple[str, str]: A tuple containing the field ID and option ID.
        """
        field_id = None
        option_id = None
        project_fields = self.get_project_fields()
        for field in project_fields:
            if field["name"] != "Status":
                continue
            field_id = field["id"]
            for option in field["options"]:
                if option["name"] == field_name:
                    option_id = option["id"]
                    break
        if not field_id or not option_id:
            raise ValueError(f"Field '{field_name}' not found in project fields.")

        return field_id, option_id

    def set_field_option(self, item_id, field_id, option_id):
        """
        Set the option of a project item for a single-select field.

        Args:
            item_id (str): The ID of the project item to update.
            field_id (str): The ID of the field to update.
            option_id (str): The ID of the option to set.
        """

        mutation = """
        mutation($projectId: ID!, $itemId: ID!, $fieldId: ID!, $optionId: String!) {
          updateProjectV2ItemFieldValue(
            input: {
              projectId: $projectId
              itemId: $itemId
              fieldId: $fieldId
              value: { singleSelectOptionId: $optionId }
            }
          ) {
            projectV2Item {
              id
            }
          }
        }
        """
        variables = {
            "projectId": self.project_node_id,
            "itemId": item_id,
            "fieldId": field_id,
            "optionId": option_id,
        }
        return self.run_graphql(mutation, variables)

    @functools.lru_cache(maxsize=1)
    def get_project_fields(self) -> list[dict]:
        """
        Get the available fields in the project.
        This method caches the result to avoid multiple API calls.

        Returns:
            list[dict]: A list of fields in the project.
        """

        query = """
    query($projectId: ID!) {
      node(id: $projectId) {
        ... on ProjectV2 {
          fields(first: 50) {
            nodes {
              ... on ProjectV2SingleSelectField {
                id
                name
                options {
                  id
                  name
                }
              }
            }
          }
        }
      }
    }
    """
        variables = {"projectId": self.project_node_id}
        resp = self.run_graphql(query, variables)
        return list(filter(bool, resp["data"]["node"]["fields"]["nodes"]))

    def get_pull_request_linked_issues(self, pr_number: int) -> list[dict]:
        """
        Get the linked issues of a pull request.

        Args:
            pr_number (int): The pull request number.

        Returns:
            list[dict]: A list of linked issues.
        """
        query = """
    query($number: Int!, $owner: String!, $repo: String!) {
      repository(owner: $owner, name: $repo) {
        pullRequest(number: $number) {
          id
          closingIssuesReferences(first: 50) {
            edges {
              node {
                id
                body
                number
                title
              }
            }
          }
        }
      }
    }
    """
        variables = {
            "number": pr_number,
            "owner": self.gh_config.organization,
            "repo": self.gh_config.repository,
        }
        resp = self.run_graphql(query, variables)
        edges = resp["data"]["repository"]["pullRequest"]["closingIssuesReferences"]["edges"]
        return [edge["node"] for edge in edges if edge.get("node")]


def main():
    # GitHub settings
    token = os.getenv("TOKEN")
    org = os.getenv("ORG")
    repo = os.getenv("REPO")
    project_number = os.getenv("PROJECT_NUMBER")
    pr_number = os.getenv("PR_NUMBER")

    if not token:
        raise ValueError("GitHub token is not set. Please set the TOKEN environment variable.")
    if not org:
        raise ValueError("GitHub organization is not set. Please set the ORG environment variable.")
    if not repo:
        raise ValueError("GitHub repository is not set. Please set the REPO environment variable.")
    if not project_number:
        raise ValueError(
            "GitHub project number is not set. Please set the PROJECT_NUMBER environment variable."
        )
    if not pr_number:
        raise ValueError(
            "Pull request number is not set. Please set the PR_NUMBER environment variable."
        )

    project_number = int(project_number)
    pr_number = int(pr_number)

    gh_config = GHConfig(
        token=token,
        organization=org,
        repository=repo,
        project_number=project_number,
        graphql_url="https://api.github.com/graphql",
        rest_url=f"https://api.github.com/repos/{org}/{repo}/issues",
        headers={"Authorization": f"Bearer {token}", "Accept": "application/vnd.github+json"},
    )
    project_item_handler = ProjectItemHandler(gh_config=gh_config)

    # Get PR info
    pr = project_item_handler.repo.get_pull(pr_number)

    # Get the linked issues of the pull request
    linked_issues = project_item_handler.get_pull_request_linked_issues(pr_number=pr_number)
    print(f"Linked issues: {linked_issues}")

    target_status = "In Development" if pr.draft else "Ready For Review"
    print(f"Target status: {target_status}")
    for issue in linked_issues:
        project_item_handler.set_issue_status(issue_number=issue["number"], status=target_status)


if __name__ == "__main__":
    main()



import threading
import time

import numpy as np
import pytest
from bec_lib import messages
from ophyd import Device

from ophyd_devices.utils.bec_signals import (
    BECMessageSignal,
    DynamicSignal,
    FileEventSignal,
    PreviewSignal,
    ProgressSignal,
)
from ophyd_devices.utils.psi_device_base_utils import (
    FileHandler,
    TaskHandler,
    TaskKilledError,
    TaskState,
    TaskStatus,
)

# pylint: disable=protected-access
# pylint: disable=redefined-outer-name

##########################################
#########  Test Task Handler  ############
##########################################


@pytest.fixture
def file_handler():
    """Fixture for FileHandler"""
    yield FileHandler()


@pytest.fixture
def device():
    """Fixture for Device"""
    yield Device(name="device")


@pytest.fixture
def task_handler(device):
    """Fixture for TaskHandler"""
    yield TaskHandler(parent=device)


def test_utils_file_handler_has_full_path(file_handler):
    """Ensure that file_handler has a get_full_path method"""
    assert hasattr(file_handler, "get_full_path")


def test_utils_task_status(device):
    """Test TaskStatus creation"""
    status = TaskStatus(device=device)
    assert status.device.name == "device"
    assert status.state == "not_started"
    assert status.task_id == status._task_id
    status.state = "running"
    assert status.state == TaskState.RUNNING
    status.state = TaskState.COMPLETED
    assert status.state == "completed"


def test_utils_task_handler_submit_task_with_args(task_handler):
    """Ensure that task_handler has a submit_task method"""

    def my_task(input_arg: bool, input_kwarg: bool = False):
        if input_kwarg is True:
            raise ValueError("input_kwarg is True")
        if input_arg is True:
            return True
        return False

    # This should fail
    with pytest.raises(TypeError):
        status = task_handler.submit_task(my_task)
        status.wait()
    # This should pass

    task_stopped = threading.Event()

    def finished_cb():
        task_stopped.set()

    status = task_handler.submit_task(
        my_task, task_args=(True,), task_kwargs={"input_kwarg": False}
    )
    status.add_callback(finished_cb)
    task_stopped.wait()
    assert status.done is True
    assert status.state == TaskState.COMPLETED
    # This should fail
    task_stopped = threading.Event()
    status = task_handler.submit_task(my_task, task_args=(True,), task_kwargs={"input_kwarg": True})
    with pytest.raises(ValueError):
        status.wait()
    assert status.state == TaskState.ERROR
    assert status.done is True
    assert status.exception().__class__ == ValueError


@pytest.mark.timeout(100)
def test_utils_task_handler_task_killed(task_handler):
    """Ensure that task_handler has a submit_task method"""
    # No tasks should be running
    assert len(task_handler._tasks) == 0
    event = threading.Event()
    task_stopped = threading.Event()
    task_started = threading.Event()

    def finished_cb():
        task_stopped.set()

    def my_wait_task():
        task_started.set()
        for _ in range(100):
            event.wait(timeout=0.1)

    # Create task
    status = task_handler.submit_task(my_wait_task, run=False)
    status.add_callback(finished_cb)
    assert status.state == TaskState.NOT_STARTED
    # Start task
    task_handler.start_task(status)
    task_started.wait()
    assert status.state == TaskState.RUNNING
    # Stop task
    task_handler.kill_task(status)
    task_stopped.wait()
    assert status.state == TaskState.KILLED
    assert status.exception().__class__ == TaskKilledError


@pytest.mark.timeout(100)
def test_utils_task_handler_task_successful(task_handler):
    """Ensure that the task handler runs a successful task"""
    assert len(task_handler._tasks) == 0
    event = threading.Event()
    task_stopped = threading.Event()
    task_started = threading.Event()

    def finished_cb():
        task_stopped.set()

    def my_wait_task():
        task_started.set()
        for _ in range(100):
            ret = event.wait(timeout=0.1)
            if ret is True:
                break

    status = task_handler.submit_task(my_wait_task, run=False)
    status.add_callback(finished_cb)
    task_handler.start_task(status)
    task_started.wait()
    assert status.state == TaskState.RUNNING
    event.set()
    task_stopped.wait()
    assert status.state == TaskState.COMPLETED


def test_utils_task_handler_shutdown(task_handler):
    """Test to shutdown the handler"""

    task_completed_cb1 = threading.Event()
    task_completed_cb2 = threading.Event()

    def finished_cb1():
        task_completed_cb1.set()

    def finished_cb2():
        task_completed_cb2.set()

    def cb1():
        for _ in range(1000):
            time.sleep(0.2)

    def cb2():
        for _ in range(1000):
            time.sleep(0.2)

    status1 = task_handler.submit_task(cb1)
    status1.add_callback(finished_cb1)
    status2 = task_handler.submit_task(cb2)
    status2.add_callback(finished_cb2)
    assert len(task_handler._tasks) == 2
    assert status1.state == TaskState.RUNNING
    assert status2.state == TaskState.RUNNING
    task_handler.shutdown()
    task_completed_cb1.wait()
    task_completed_cb2.wait()
    assert len(task_handler._tasks) == 0
    assert status1.state == TaskState.KILLED
    assert status2.state == TaskState.KILLED
    assert status1.exception().__class__ == TaskKilledError


##########################################
#########  Test PSI cusomt signals  ######
##########################################


def test_utils_bec_message_signal():
    """Test BECMessageSignal"""
    dev = Device(name="device")
    signal = BECMessageSignal(
        name="bec_message_signal",
        bec_message_type=messages.GUIInstructionMessage,
        value=None,
        parent=dev,
    )
    assert signal.parent == dev
    assert signal._bec_message_type == messages.GUIInstructionMessage
    assert signal._readback is None
    assert signal.name == "bec_message_signal"
    assert signal.describe() == {
        "bec_message_signal": {
            "source": "BECMessageSignal:bec_message_signal",
            "dtype": "GUIInstructionMessage",
            "shape": [],
            "signal_info": {
                "data_type": "raw",
                "saved": True,
                "ndim": 0,
                "scope": "scan",
                "role": "main",
                "enabled": True,
                "rpc_access": False,
                "signals": [("bec_message_signal", 5)],
                "signal_metadata": None,
            },
        }
    }
    # Put works with Message
    msg = messages.GUIInstructionMessage(action="image", parameter={"gui_id": "test"})
    signal.put(msg)
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # set works with dict, should call put
    msg_dict = {"action": "image", "parameter": {"gui_id": "test"}}
    status = signal.set(msg_dict)
    assert status.done is True
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # Put fails with wrong type
    with pytest.raises(ValueError):
        signal.put("wrong_type")
    # Put fails with wrong dict
    with pytest.raises(ValueError):
        signal.put({"wrong_key": "wrong_value"})


def test_utils_dynamic_signal():
    """Test DynamicSignal"""
    dev = Device(name="device")
    signal = DynamicSignal(name="dynamic_signal", signals=["sig1", "sig2"], value=None, parent=dev)
    assert signal.parent == dev
    assert signal._bec_message_type == messages.DeviceMessage
    assert signal._readback is None
    assert signal.name == "dynamic_signal"
    assert signal.signals == [("sig1", 1), ("sig2", 1)]
    assert signal.describe() == {
        "dynamic_signal": {
            "source": "BECMessageSignal:dynamic_signal",
            "dtype": "DeviceMessage",
            "shape": [],
            "signal_info": {
                "data_type": "raw",
                "saved": True,
                "ndim": 1,
                "scope": "scan",
                "role": "main",
                "enabled": True,
                "rpc_access": False,
                "signals": [("sig1", 1), ("sig2", 1)],
                "signal_metadata": None,
            },
        }
    }

    # Put works with Message
    msg_dict = {"dynamic_signal_sig1": {"value": 1}, "dynamic_signal_sig2": {"value": 2}}
    msg = messages.DeviceMessage(signals=msg_dict)
    signal.put(msg)
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # Set works with dict
    status = signal.set(msg_dict)
    assert status.done is True
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # Put fails with wrong type
    with pytest.raises(ValueError):
        signal.put("wrong_type")
    # Put fails with wrong dict
    with pytest.raises(ValueError):
        signal.put({"wrong_key": "wrong_value"})


def test_utils_file_event_signal():
    """Test FileEventSignal"""
    dev = Device(name="device")
    signal = FileEventSignal(name="file_event_signal", value=None, parent=dev)
    assert signal.parent == dev
    assert signal._bec_message_type == messages.FileMessage
    assert signal._readback is None
    assert signal.name == "file_event_signal"
    assert signal.describe() == {
        "file_event_signal": {
            "source": "BECMessageSignal:file_event_signal",
            "dtype": "FileMessage",
            "shape": [],
            "signal_info": {
                "data_type": "raw",
                "saved": False,
                "ndim": 0,
                "scope": "scan",
                "role": "file event",
                "enabled": True,
                "rpc_access": False,
                "signals": [("file_event_signal", 5)],
                "signal_metadata": None,
            },
        }
    }

    # Test put works with FileMessage
    msg_dict = {"file_path": "/path/to/another/file.txt", "done": False, "successful": True}
    msg = messages.FileMessage(**msg_dict)
    signal.put(msg)
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # Test put works with dict
    signal.put(msg_dict)
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # Test set with kwargs, should call put
    status = signal.set(file_path="/path/to/another/file.txt", done=False, successful=True)
    assert status.done is True
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # Test put fails with wrong type
    with pytest.raises(ValueError):
        signal.put(1)
    # Test put fails with wrong dict
    with pytest.raises(ValueError):
        signal.put({"wrong_key": "wrong_value"})


def test_utils_preview_1d_signal():
    """Test Preview1DSignal"""
    dev = Device(name="device")
    signal = PreviewSignal(name="preview_1d_signal", ndim=1, value=None, parent=dev)
    assert signal.ndim == 1
    assert signal.parent == dev
    assert signal._bec_message_type == messages.DevicePreviewMessage
    assert signal._readback is None
    assert signal.name == "preview_1d_signal"
    assert signal.describe() == {
        "preview_1d_signal": {
            "source": "BECMessageSignal:preview_1d_signal",
            "dtype": "DevicePreviewMessage",
            "shape": [],
            "signal_info": {
                "data_type": "raw",
                "saved": False,
                "ndim": 1,
                "scope": "scan",
                "role": "preview",
                "enabled": True,
                "rpc_access": False,
                "signals": [("preview_1d_signal", 5)],
                "signal_metadata": {"num_rotation_90": 0, "transpose": False},
            },
        }
    }
    # Put works with Message
    msg_dict = {"device": dev.name, "data": np.array([1, 2, 3]), "signal": "preview_1d_signal"}
    msg = messages.DevicePreviewMessage(**msg_dict)
    signal.put(msg)
    reading = signal.read()
    assert reading[signal.name]["value"].model_dump(exclude="timestamp") == msg.model_dump(
        exclude="timestamp"
    )
    # Put works with dict
    signal.put(msg_dict)
    reading = signal.read()
    assert reading[signal.name]["value"].model_dump(exclude="timestamp") == msg.model_dump(
        exclude="timestamp"
    )
    # Put works with value
    status = signal.set(msg_dict["data"])
    assert status.done is True
    reading = signal.read()
    assert reading[signal.name]["value"].model_dump(exclude="timestamp") == msg.model_dump(
        exclude="timestamp"
    )
    # Put works with value
    signal.put(msg_dict["data"])
    reading = signal.read()
    assert reading[signal.name]["value"].model_dump(exclude="timestamp") == msg.model_dump(
        exclude="timestamp"
    )
    # Put fails with wrong type
    with pytest.raises(ValueError):
        signal.put(1)
    # Put fails with wrong dict
    with pytest.raises(ValueError):
        signal.put({"wrong_key": "wrong_value"})


def test_utils_preview_2d_signal():
    """Test Preview2DSignal"""
    dev = Device(name="device")
    signal = PreviewSignal(name="preview_2d_signal", ndim=2, value=None, parent=dev)
    assert signal.ndim == 2
    assert signal.parent == dev
    assert signal._bec_message_type == messages.DevicePreviewMessage
    assert signal._readback is None
    assert signal.name == "preview_2d_signal"
    assert signal.describe() == {
        "preview_2d_signal": {
            "source": "BECMessageSignal:preview_2d_signal",
            "dtype": "DevicePreviewMessage",
            "shape": [],
            "signal_info": {
                "data_type": "raw",
                "saved": False,
                "ndim": 2,
                "scope": "scan",
                "role": "preview",
                "enabled": True,
                "rpc_access": False,
                "signals": [("preview_2d_signal", 5)],
                "signal_metadata": {"num_rotation_90": 0, "transpose": False},
            },
        }
    }
    # Put works with Message
    msg_dict = {
        "device": dev.name,
        "data": np.array([[1, 2, 3], [4, 5, 6]]),
        "signal": "preview_2d_signal",
    }
    msg = messages.DevicePreviewMessage(**msg_dict)
    signal.put(msg)
    reading = signal.read()
    assert reading[signal.name]["value"].model_dump(exclude="timestamp") == msg.model_dump(
        exclude="timestamp"
    )
    # Put works with dict
    signal.put(msg_dict)
    reading = signal.read()
    assert reading[signal.name]["value"].model_dump(exclude="timestamp") == msg.model_dump(
        exclude="timestamp"
    )
    # Put works with value
    status = signal.set(msg_dict["data"])
    assert status.done is True
    reading = signal.read()
    assert reading[signal.name]["value"].model_dump(exclude="timestamp") == msg.model_dump(
        exclude="timestamp"
    )
    # Put works with value
    signal.put(msg_dict["data"])
    reading = signal.read()
    assert reading[signal.name]["value"].model_dump(exclude="timestamp") == msg.model_dump(
        exclude="timestamp"
    )
    # Put fails with wrong type
    with pytest.raises(ValueError):
        signal.put(1)
    # Put fails with wrong dict
    with pytest.raises(ValueError):
        signal.put({"wrong_key": "wrong_value"})


def test_utils_progress_signal():
    """Test ProgressSignal"""
    dev = Device(name="device")
    signal = ProgressSignal(name="progress_signal", value=None, parent=dev)
    assert signal.parent == dev
    assert signal._bec_message_type == messages.ProgressMessage
    assert signal._readback is None
    assert signal.name == "progress_signal"
    assert signal.describe() == {
        "progress_signal": {
            "source": "BECMessageSignal:progress_signal",
            "dtype": "ProgressMessage",
            "shape": [],
            "signal_info": {
                "data_type": "raw",
                "saved": False,
                "ndim": 0,
                "scope": "scan",
                "role": "progress",
                "enabled": True,
                "rpc_access": False,
                "signals": [("progress_signal", 5)],
                "signal_metadata": None,
            },
        }
    }
    # Put works with Message
    msg = messages.ProgressMessage(value=1, max_value=10, done=False)
    signal.put(msg)
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # Put works with dict
    msg_dict = {"value": 1, "max_value": 10, "done": False}
    signal.put(msg_dict)
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # Works with kwargs
    status = signal.set(value=1, max_value=10, done=False)
    assert status.done is True
    reading = signal.read()
    assert reading[signal.name]["value"] == msg
    # Put fails with wrong type
    with pytest.raises(ValueError):
        signal.put(1)
    # Put fails with wrong dict
    with pytest.raises(ValueError):
        signal.put({"wrong_key": "wrong_value"})



import time

from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as Dcpt
from ophyd import EpicsSignalRO


class SLSOperatorMessages(Device):
    SUB_VALUE = "value"
    _default_sub = SUB_VALUE
    messages_info = {
        f"message{i}": (EpicsSignalRO, f"ACOAU-ACCU:OP-MSG{i}", {}) for i in range(1, 6)
    }
    messages = Dcpt(messages_info)
    date_info = {f"message{i}": (EpicsSignalRO, f"ACOAU-ACCU:OP-DATE{i}", {}) for i in range(1, 6)}
    date = Dcpt(date_info)

    def __init__(self, prefix="", *, name, **kwargs):
        super().__init__(prefix, name=name, **kwargs)
        self.messages.message1.subscribe(self._emit_value)

    def _emit_value(self, **kwargs):
        timestamp = kwargs.pop("timestamp", time.time())
        self.wait_for_connection()
        self._run_subs(sub_type=self.SUB_VALUE, timestamp=timestamp, obj=self)


class SLSInfo(Device):
    SUB_VALUE = "value"
    _default_sub = SUB_VALUE
    # eh_t02_avg_temperature = Cpt(EpicsSignalRO, "ILUUL-02AV:TEMP", auto_monitor=True)
    # eh_t02_temperature_t0204_axis_16 = Cpt(
    #     EpicsSignalRO, "ILUUL-0200-EB104:TEMP", auto_monitor=True
    # )
    # eh_t02_temperature_t0205_axis_18 = Cpt(
    #     EpicsSignalRO, "ILUUL-0200-EB105:TEMP", auto_monitor=True
    # )
    injection_mode = Cpt(EpicsSignalRO, "ALIRF-GUN:INJ-MODE", auto_monitor=True, string=True)
    current_threshold = Cpt(EpicsSignalRO, "ALIRF-GUN:CUR-LOWLIM", auto_monitor=True)
    current_deadband = Cpt(EpicsSignalRO, "ALIRF-GUN:CUR-DBAND", auto_monitor=True)
    filling_pattern = Cpt(EpicsSignalRO, "ACORF-FILL:PAT-SELECT", auto_monitor=True, string=True)
    filling_life_time = Cpt(EpicsSignalRO, "ARIDI-PCT:TAU-HOUR", auto_monitor=True)
    orbit_feedback_mode = Cpt(EpicsSignalRO, "ARIDI-BPM:OFB-MODE", auto_monitor=True, string=True)
    fast_orbit_feedback = Cpt(
        EpicsSignalRO, "ARIDI-BPM:FOFBSTATUS-G", auto_monitor=True, string=True
    )
    ring_current = Cpt(EpicsSignalRO, "ARIDI-PCT:CURRENT", auto_monitor=True)
    machine_status = Cpt(EpicsSignalRO, "ACOAU-ACCU:OP-MODE", auto_monitor=True, string=True)
    crane_usage = Cpt(
        EpicsSignalRO, "IBWKR-0101-QH10003:D01_H_D-WA", auto_monitor=True, string=True
    )

    def __init__(self, prefix="", *, name, **kwargs):
        super().__init__(prefix, name=name, **kwargs)
        self.ring_current.subscribe(self._emit_value)

    def _emit_value(self, **kwargs):
        timestamp = kwargs.pop("timestamp", time.time())
        self.wait_for_connection()
        self._run_subs(sub_type=self.SUB_VALUE, timestamp=timestamp, obj=self)



"""AreaDetector Devices"""

# isort: skip_file
from ophyd import EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV
from ophyd.areadetector import ADComponent as ADCpt
from ophyd.areadetector.cam import (
    CamBase as _CamBase,
    FileBase as _FileBase,
    Andor3DetectorCam as _Andor3DetectorCam,
    PilatusDetectorCam as _PilatusDetectorCam,
    EigerDetectorCam as _EigerDetectorCam,
    ProsilicaDetectorCam as _ProsilicaDetectorCam,
    SimDetectorCam as _SimDetectorCam,
    URLDetectorCam as _URLDetectorCam,
)

__all__ = [
    "CamBase",
    "FileBase",
    "Andor3DetectorCam",
    "EigerDetectorCam",
    "PilatusDetectorCam",
    "ProsilicaDetectorCam",
    "URLDetectorCam",
    "AravisDetectorCam",
    "PylonDetectorCam",
    "VimbaDetectorCam",
]


class CamBase(_CamBase):
    """
    Base class for all camera drivers.
    """

    pool_max_buffers = None


class FileBase(_FileBase):
    """
    File saving parameters.

    It is not meant to be used directly, but rather through inheritance by camera drivers
    with file saving support, e.g. PilatusDetectorCam and SLSDetectorCam
    """

    file_number_sync = None
    file_number_write = None


class Andor3DetectorCam(CamBase, _Andor3DetectorCam):
    """
    ADAndor3 driver, https://github.com/areaDetector/ADAndor3

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(Andor3DetectorCam, 'cam1:')

    """

    gate_mode = ADCpt(EpicsSignalWithRBV, "GateMode")
    insertion_delay = ADCpt(EpicsSignalWithRBV, "InsertionDelay")
    mcp_gain = ADCpt(EpicsSignalWithRBV, "MCPGain")
    mcp_intelligate = ADCpt(EpicsSignalWithRBV, "MCPIntelligate")


class EigerDetectorCam(CamBase, _EigerDetectorCam):
    """
    ADEiger driver, https://github.com/areaDetector/ADEiger

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(EigerDetectorCam, 'cam1:')

    """


class PilatusDetectorCam(CamBase, FileBase, _PilatusDetectorCam):
    """
    ADPilatus driver, https://github.com/areaDetector/ADPilatus

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(PilatusDetectorCam, 'cam1:')

    """


class ProsilicaDetectorCam(CamBase, _ProsilicaDetectorCam):
    """
    ADProsilica driver, https://github.com/areaDetector/ADProsilica

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(ProsilicaDetectorCam, 'cam1:')

    """


class SimDetectorCam(CamBase, _SimDetectorCam):
    """
    ADSimDetector driver, https://github.com/areaDetector/ADSimDetector

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(SimDetectorCam, 'cam1:')

    """


class URLDetectorCam(CamBase, _URLDetectorCam):
    """
    ADURL driver, https://github.com/areaDetector/ADURL

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(AravisDetectorCam, 'cam1:')

    """


class GenICam(CamBase):
    """
    ADGenICam driver, https://github.com/areaDetector/ADGenICam

    It is the base class for GenICam drivers and not meant to be used directly.
    """

    frame_rate = ADCpt(EpicsSignalWithRBV, "FrameRate")
    frame_rate_enable = ADCpt(EpicsSignalWithRBV, "FrameRateEnable")
    trigger_source = ADCpt(EpicsSignalWithRBV, "TriggerSource")
    trigger_overlap = ADCpt(EpicsSignalWithRBV, "TriggerOverlap")
    trigger_software = ADCpt(EpicsSignal, "TriggerSoftware")
    exposure_mode = ADCpt(EpicsSignalWithRBV, "ExposureMode")
    exposure_auto = ADCpt(EpicsSignalWithRBV, "ExposureAuto")
    gain_auto = ADCpt(EpicsSignalWithRBV, "GainAuto")
    pixel_format = ADCpt(EpicsSignalWithRBV, "PixelFormat")


class AravisDetectorCam(GenICam):
    """
    ADAravis driver, https://github.com/areaDetector/ADAravis

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(AravisDetectorCam, 'cam1:')

    """

    ar_convert_pixel_format = ADCpt(EpicsSignalWithRBV, "ARConvertPixelFormat")
    ar_shift_dir = ADCpt(EpicsSignalWithRBV, "ARShiftDir")
    ar_shift_bits = ADCpt(EpicsSignalWithRBV, "ARShiftBits")


class VimbaDetectorCam(GenICam):
    """
    ADVimba driver, https://github.com/areaDetector/ADVimba

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(VimbaDetectorCam, 'cam1:')

    """

    time_stamp_mode = ADCpt(EpicsSignalWithRBV, "TimeStampMode")
    unique_id_mode = ADCpt(EpicsSignalWithRBV, "UniqueIdMode")
    convert_pixel_format = ADCpt(EpicsSignalWithRBV, "ConvertPixelFormat")


class PylonDetectorCam(GenICam):
    """
    ADPylon driver, https://github.com/areaDetector/ADPylon

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(PylonDetectorCam, 'cam1:')

    """

    time_stamp_mode = ADCpt(EpicsSignalWithRBV, "TimeStampMode")
    unique_id_mode = ADCpt(EpicsSignalWithRBV, "UniqueIdMode")
    convert_pixel_format = ADCpt(EpicsSignalWithRBV, "ConvertPixelFormat")
    convert_bit_align = ADCpt(EpicsSignalWithRBV, "ConvertBitAlign")
    convert_shift_bits = ADCpt(EpicsSignalWithRBV, "ConvertShiftBits")


class SLSDetectorCam(CamBase, FileBase):
    """
    slsDetector driver, https://github.com/paulscherrerinstitute/slsDetector

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(SLSDetectorCam, 'cam1:')

    """

    detector_type = ADCpt(EpicsSignalRO, "DetectorType_RBV")
    setting = ADCpt(EpicsSignalWithRBV, "Setting")
    delay_time = ADCpt(EpicsSignalWithRBV, "DelayTime")
    threshold_energy = ADCpt(EpicsSignalWithRBV, "ThresholdEnergy")
    enable_trimbits = ADCpt(EpicsSignalWithRBV, "Trimbits")
    bit_depth = ADCpt(EpicsSignalWithRBV, "BitDepth")
    num_gates = ADCpt(EpicsSignalWithRBV, "NumGates")
    num_cycles = num_images = ADCpt(EpicsSignalWithRBV, "NumCycles")
    num_frames = ADCpt(EpicsSignalWithRBV, "NumFrames")
    trigger_mode = timing_mode = ADCpt(EpicsSignalWithRBV, "TimingMode")
    trigger_software = ADCpt(EpicsSignal, "TriggerSoftware")
    high_voltage = ADCpt(EpicsSignalWithRBV, "HighVoltage")
    # Receiver and data callback
    receiver_mode = ADCpt(EpicsSignalWithRBV, "ReceiverMode")
    receiver_stream = ADCpt(EpicsSignalWithRBV, "ReceiverStream")
    enable_data = ADCpt(EpicsSignalWithRBV, "UseDataCallback")
    missed_packets = ADCpt(EpicsSignalRO, "ReceiverMissedPackets_RBV")
    # Direct settings access
    setup_file = ADCpt(EpicsSignal, "SetupFile")
    load_setup = ADCpt(EpicsSignal, "LoadSetup")
    command = ADCpt(EpicsSignal, "Command")
    # Mythen 3
    counter_mask = ADCpt(EpicsSignalWithRBV, "CounterMask")
    counter1_threshold = ADCpt(EpicsSignalWithRBV, "Counter1Threshold")
    counter2_threshold = ADCpt(EpicsSignalWithRBV, "Counter2Threshold")
    counter3_threshold = ADCpt(EpicsSignalWithRBV, "Counter3Threshold")
    gate1_delay = ADCpt(EpicsSignalWithRBV, "Gate1Delay")
    gate1_width = ADCpt(EpicsSignalWithRBV, "Gate1Width")
    gate2_delay = ADCpt(EpicsSignalWithRBV, "Gate2Delay")
    gate2_width = ADCpt(EpicsSignalWithRBV, "Gate2Width")
    gate3_delay = ADCpt(EpicsSignalWithRBV, "Gate3Delay")
    gate3_width = ADCpt(EpicsSignalWithRBV, "Gate3Width")
    # Moench
    json_frame_mode = ADCpt(EpicsSignalWithRBV, "JsonFrameMode")
    json_detector_mode = ADCpt(EpicsSignalWithRBV, "JsonDetectorMode")


class ASItpxCam(CamBase):
    """
    ASI Timepix detector driver, https://github.com/paulscherrerinstitute/ADASItpx

    ::

        from ophyd import Component as Cpt

        class MyDetector(ADBase):
            cam = Cpt(ASItpxCam, 'cam1:')

    """

    trigger_source = ADCpt(EpicsSignalWithRBV, "TriggerSource")
    trigger_polarity = ADCpt(EpicsSignalWithRBV, "TriggerPolarity")
    trigger_delay = ADCpt(EpicsSignalWithRBV, "TriggerDelay")
    exposure_mode = ADCpt(EpicsSignalWithRBV, "ExposureMode")
    trigger_software = ADCpt(EpicsSignal, "TriggerSoftware")

    raw_enable = ADCpt(EpicsSignalWithRBV, "RawEnable")
    raw_file_path = ADCpt(EpicsSignalWithRBV, "RawFilePath")
    raw_file_template = ADCpt(EpicsSignalWithRBV, "RawFileTemplate")

    # pixel mode for the 2D image
    pixel_mode = ADCpt(EpicsSignalWithRBV, "PixelMode")

    image_enable = ADCpt(EpicsSignalWithRBV, "ImageEnable")
    image_file_path = ADCpt(EpicsSignalWithRBV, "ImageFilePath")
    image_file_template = ADCpt(EpicsSignalWithRBV, "ImageFileTemplate")

    preview_enable = ADCpt(EpicsSignalWithRBV, "PreviewEnable")
    preview_period = ADCpt(EpicsSignalWithRBV, "PreviewPeriod")



from abc import ABC, abstractmethod
from collections import defaultdict

from ophyd_devices.utils.bec_device_base import BECDeviceBase


class DeviceProxy(BECDeviceBase, ABC):
    """DeviceProxy class inherits from BECDeviceBase.

    It is an abstract class that is meant to be used as a base class for all device proxies.
    The minimum requirement for a device proxy is to implement the _compute method.
    """

    def __init__(self, name, *args, device_manager=None, **kwargs):
        self.name = name
        self.device_manager = device_manager
        self.config = None
        self._lookup = defaultdict(dict)
        super().__init__(name, *args, device_manager=device_manager, **kwargs)
        self._signals = dict()

    @property
    def lookup(self):
        """lookup property"""
        return self._lookup

    @lookup.setter
    def lookup(self, update: dict) -> None:
        """lookup setter"""
        self._lookup.update(update)

    def _update_device_config(self, config: dict) -> None:
        """
        BEC will call this method on every object upon initializing devices to pass over the deviceConfig
        from the config file. It can be conveniently be used to hand over initial parameters to the device.

        Args:
            config (dict): Config dictionary.
        """
        self.config = config
        self._compile_lookup()

    def _compile_lookup(self):
        """Compile the lookup table for the device."""
        for device_name in self.config.keys():
            self._lookup[device_name] = {
                "method": self._compute,
                "signal_name": self.config[device_name]["signal_name"],
                "args": (device_name,),
                "kwargs": {},
            }

    @abstractmethod
    def _compute(self, device_name: str, *args, **kwargs) -> any:
        """
        The purpose of this method is to compute the readback value for the signal of the device
        that this proxy is attached to. This method is meant to be overriden by the user.
        P

        Args:
            device_name (str): Name of the device.

        Returns:
        """



import getpass

from bec_lib import bec_logger, messages
from bec_lib.devicemanager import DeviceManagerBase
from bec_lib.endpoints import MessageEndpoints

from ophyd_devices.utils.bec_utils import DMMock

logger = bec_logger.logger


class BECInfoMsgMock:
    """Mock BECInfoMsg class

    This class is used for mocking BECInfoMsg for testing purposes
    """

    def __init__(
        self,
        mockrid: str = "mockrid1111",
        mockqueueid: str = "mockqueue_id111",
        scan_number: int = 1,
        exp_time: float = 15e-3,
        num_points: int = 500,
        readout_time: float = 3e-3,
        scan_type: str = "fly",
        num_lines: int = 1,
        frames_per_trigger: int = 1,
    ) -> None:
        self.mockrid = mockrid
        self.mockqueueid = mockqueueid
        self.scan_number = scan_number
        self.exp_time = exp_time
        self.num_points = num_points
        self.readout_time = readout_time
        self.scan_type = scan_type
        self.num_lines = num_lines
        self.frames_per_trigger = frames_per_trigger

    def get_bec_info_msg(self) -> dict:
        """Get BECInfoMsg object"""
        info_msg = {
            "RID": self.mockrid,
            "queue_id": self.mockqueueid,
            "scan_number": self.scan_number,
            "exp_time": self.exp_time,
            "num_points": self.num_points,
            "readout_time": self.readout_time,
            "scan_type": self.scan_type,
            "num_lines": self.exp_time,
            "frames_per_trigger": self.frames_per_trigger,
        }

        return info_msg


class BecScaninfoMixin:
    """BecScaninfoMixin class

    Args:
        device_manager (DeviceManagerBase): DeviceManagerBase object
        sim_mode (bool): Simulation mode flag
        bec_info_msg (dict): BECInfoMsg object
    Returns:
        BecScaninfoMixin: BecScaninfoMixin object
    """

    def __init__(self, device_manager: DeviceManagerBase = None, bec_info_msg=None) -> None:
        self.sim_mode = bool(isinstance(device_manager, DMMock))
        self.device_manager = device_manager
        self.scan_msg = None
        self.scan_id = None
        if bec_info_msg is None:
            infomsgmock = BECInfoMsgMock()
            self.bec_info_msg = infomsgmock.get_bec_info_msg()
        else:
            self.bec_info_msg = bec_info_msg

        self.metadata = None
        self.scan_id = None
        self.scan_number = None
        self.exp_time = None
        self.frames_per_trigger = None
        self.num_points = None
        self.scan_type = None

    def get_bec_info_msg(self) -> None:
        """Get BECInfoMsg object"""
        return self.bec_info_msg

    def change_config(self, bec_info_msg: dict) -> None:
        """Change BECInfoMsg object"""
        self.bec_info_msg = bec_info_msg

    def _get_current_scan_msg(self) -> messages.ScanStatusMessage:
        """Get current scan message

        Returns:
            messages.ScanStatusMessage: messages.ScanStatusMessage object
        """
        if not self.sim_mode:
            msg = self.device_manager.connector.get(MessageEndpoints.scan_status())
            if not isinstance(msg, messages.ScanStatusMessage):
                return None
            return msg

        return messages.ScanStatusMessage(scan_id="1", status="open", info=self.bec_info_msg)

    def get_username(self) -> str:
        """Get username"""
        if self.sim_mode:
            return getpass.getuser()

        msg = self.device_manager.connector.get(MessageEndpoints.account())
        if msg:
            return msg
        return getpass.getuser()

    def load_scan_metadata(self) -> None:
        """Load scan metadata

        This function loads scan metadata from the current scan message
        """
        self.scan_msg = scan_msg = self._get_current_scan_msg()
        try:
            logger.info(f"Received scan msg for {self.scan_msg.content['scan_id']}")
            self.metadata = {
                "scan_id": scan_msg.content["scan_id"],
                "RID": scan_msg.content["info"]["RID"],
                "queue_id": scan_msg.content["info"]["queue_id"],
            }
            self.scan_id = scan_msg.content["scan_id"]
            self.scan_number = scan_msg.content["info"]["scan_number"]
            self.exp_time = scan_msg.content["info"]["exp_time"]
            self.frames_per_trigger = scan_msg.content["info"]["frames_per_trigger"]
            self.num_points = scan_msg.content["info"]["num_points"]
            self.scan_type = scan_msg.content["info"].get("scan_type", "step")
            self.readout_time = scan_msg.content["info"]["readout_time"]
        except Exception as exc:
            logger.error(f"Failed to load scan metadata: {exc}.")

        self.username = self.get_username()



"""Module for signals of the ophyd_devices simulation."""

import time
from typing import Any

import numpy as np
from bec_lib import bec_logger
from ophyd import DeviceStatus, Kind, Signal
from ophyd.utils import ReadOnlyError

from ophyd_devices.utils.bec_device_base import BECDeviceBase

logger = bec_logger.logger

# Readout precision for Setable/ReadOnlySignal signals
PRECISION = 3


class SetableSignal(Signal):
    """Setable signal for simulated devices.

    The signal will store the value in sim_state of the SimulatedData class of the parent device.
    It will also return the value from sim_state when get is called. Compared to the ReadOnlySignal,
    this signal can be written to.
    The setable signal inherits from the Signal class of ophyd, thus the class attribute needs to be
    initiated as a Component (class from ophyd).

    >>> signal = SetableSignal(name="signal", parent=parent, value=0)

    Parameters
    ----------

    name  (string)           : Name of the signal
    parent (object)          : Parent object of the signal, default none.
    value (any)              : Initial value of the signal, default 0.
    kind (int)               : Kind of the signal, default Kind.normal.
    precision (float)        : Precision of the signal, default PRECISION.
    """

    SUB_VALUE = "value"

    def __init__(
        self,
        name: str,
        *args,
        value: any = 0,
        kind: int = Kind.normal,
        precision: float = PRECISION,
        **kwargs,
    ):
        super().__init__(*args, name=name, value=value, kind=kind, **kwargs)
        self._metadata.update(connected=True, write_access=False)
        self._value = value
        self.precision = precision
        self.sim = getattr(self.parent, "sim", None)
        self._update_sim_state(value)
        self._metadata.update(write_access=True)

    def _update_sim_state(self, value: Any) -> None:
        """Update the readback value."""
        if self.sim:
            self.sim.update_sim_state(self.name, value)

    def _get_value(self) -> Any:
        """Update the timestamp of the readback value."""
        if self.sim:
            return self.sim.sim_state[self.name]["value"]
        return self._value

    def _get_timestamp(self) -> float:
        """Update the timestamp of the readback value."""
        if self.sim:
            return self.sim.sim_state[self.name]["timestamp"]
        return time.time()

    # pylint: disable=arguments-differ
    def get(self, **kwargs) -> Any:
        """Get the current position of the simulated device.

        Core function for signal.
        """
        self._value = self._get_value()
        return self._value

    # pylint: disable=arguments-differ
    def put(self, value) -> None:
        """Put the value to the simulated device.

        Core function for signal.
        """
        self.check_value(value)
        self._update_sim_state(value)
        self._value = value
        # self._run_subs(sub_type=self.SUB_VALUE, old_value=old_value, value=value, **md_for_callback)
        super().put(value)

    def set(self, value):
        """Set method"""
        self.put(value)
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def describe(self):
        """Describe the readback signal.

        Core function for signal.
        """
        res = super().describe()
        if self.precision is not None:
            res[self.name]["precision"] = self.precision
        return res

    @property
    def timestamp(self) -> float:
        """Timestamp of the readback value"""
        return self._get_timestamp()


class ReadOnlySignal(Signal):
    """Computed readback signal for simulated devices.

    The readback will be computed from a function hosted in the SimulatedData class from the parent device
    if compute_readback is True. Else, it will return the value stored int sim.sim_state directly.
    The readonly signal inherits from the Signal class of ophyd, thus the class attribute needs to be
    initiated as a Component (class from ophyd).

    >>> signal = ComputedReadOnlySignal(name="signal", parent=parent, value=0, compute_readback=True)

    Parameters
    ----------

    name  (string)           : Name of the signal
    parent (object)          : Parent object of the signal, default none.
    value (any)              : Initial value of the signal, default 0.
    kind (int)               : Kind of the signal, default Kind.normal.
    precision (float)        : Precision of the signal, default PRECISION.
    compute_readback (bool)  : Flag whether to compute readback based on function hosted in SimulatedData
                               class. If False, sim_state value will be returned, if True, new value will be computed
    """

    def __init__(
        self,
        name: str,
        *args,
        parent=None,
        value: any = 0,
        kind: int = Kind.normal,
        precision: float = PRECISION,
        compute_readback: bool = False,
        sim=None,
        **kwargs,
    ):
        super().__init__(*args, name=name, parent=parent, value=value, kind=kind, **kwargs)
        self._metadata.update(connected=True, write_access=False)
        self._value = value  # In a signal, this is self._readback
        self.precision = precision
        self.compute_readback = compute_readback
        self.sim = sim if sim is not None else getattr(self.parent, "sim", None)
        if self.sim:
            self._init_sim_state()
        self._metadata.update(write_access=False)

    def _init_sim_state(self) -> None:
        """Create the initial sim_state in the SimulatedData class of the parent device."""
        self.sim.update_sim_state(self.name, self._value)

    def _update_sim_state(self) -> None:
        """Update the readback value."""
        self.sim.compute_sim_state(signal_name=self.name, compute_readback=self.compute_readback)

    def _get_value(self) -> any:
        """Update the timestamp of the readback value."""
        return self.sim.sim_state[self.name]["value"]

    def _get_timestamp(self) -> any:
        """Update the timestamp of the readback value."""
        return self.sim.sim_state[self.name]["timestamp"]

    # pylint: disable=arguments-differ
    def get(self):
        """Get the current position of the simulated device."""
        if self.sim:
            self._update_sim_state()
            self._value = self._get_value()
            return self._value
        return np.random.rand()

    # pylint: disable=arguments-differ
    def put(self, value) -> None:
        """Put method, should raise ReadOnlyError since the signal is readonly."""
        raise ReadOnlyError(f"The signal {self.name} is readonly.")

    def describe(self):
        """Describe the readback signal.

        Core function for signal.
        """
        res = super().describe()
        if self.precision is not None:
            res[self.name]["precision"] = self.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        if self.sim:
            return self._get_timestamp()
        return time.time()



"""
Base class for all PSI ophyd device integration to ensure consistent configuration
"""

from __future__ import annotations

import inspect
import time
from typing import TYPE_CHECKING, Callable

from ophyd import Device, DeviceStatus, Staged, StatusBase

from ophyd_devices.tests.utils import get_mock_scan_info
from ophyd_devices.utils.psi_device_base_utils import FileHandler, TaskHandler

if TYPE_CHECKING:  # pragma: no cover
    from bec_lib.devicemanager import DeviceManagerBase, ScanInfo


class DeviceStoppedError(Exception):
    """Exception raised when a device is stopped"""


class PSIDeviceBase(Device):
    """
    Base class for all PSI ophyd devices to ensure consistent configuration
    and communication with BEC services.
    """

    # These are all possible subscription types that the device_manager supports
    # and automatically subscribes to
    SUB_READBACK = "readback"
    SUB_VALUE = "value"
    SUB_DONE_MOVING = "done_moving"
    SUB_MOTOR_IS_MOVING = "motor_is_moving"
    SUB_PROGRESS = "progress"
    SUB_FILE_EVENT = "file_event"
    SUB_DEVICE_MONITOR_1D = "device_monitor_1d"
    SUB_DEVICE_MONITOR_2D = "device_monitor_2d"
    _default_sub = SUB_VALUE

    def __init__(
        self,
        *,
        name: str,
        prefix: str = "",
        scan_info: ScanInfo | None = None,
        device_manager: DeviceManagerBase | None = None,
        **kwargs,
    ):
        """
        Initialize the PSI Device Base class.

        Args:
            name (str) : Name of the device
            scan_info (ScanInfo): The scan info to use.
        """
        # Make sure device_manager is not passed to super().__init__ if not specified
        # This is to avoid issues with ophyd.OphydObject.__init__ when the parent is ophyd.Device
        # and the device_manager is passed to it. This will cause a TypeError.
        sig = inspect.signature(super().__init__)
        if "device_manager" in sig.parameters:
            super().__init__(device_manager=device_manager, prefix=prefix, name=name, **kwargs)
        else:
            super().__init__(prefix=prefix, name=name, **kwargs)
        self._stopped = False
        self.task_handler = TaskHandler(parent=self)
        self.file_utils = FileHandler()
        if scan_info is None:
            scan_info = get_mock_scan_info(device=self)
        self.scan_info = scan_info
        self.on_init()

    ########################################
    # Additional Properties and Attributes #
    ########################################

    @property
    def destroyed(self) -> bool:
        """Check if the device has been destroyed."""
        return self._destroyed

    @property
    def staged(self) -> Staged:
        """Check if the device has been staged."""
        return self._staged

    @property
    def stopped(self) -> bool:
        """Check if the device has been stopped."""
        return self._stopped

    @stopped.setter
    def stopped(self, value: bool):
        self._stopped = value

    ########################################
    # Wrapper around Device class methods  #
    ########################################

    def stage(self) -> list[object] | DeviceStatus | StatusBase:  # type: ignore
        """Stage the device."""
        if self.staged != Staged.no:
            return super().stage()
        self.stopped = False
        super_staged = super().stage()
        status = self.on_stage()  # pylint: disable=assignment-from-no-return
        if isinstance(status, StatusBase):
            return status
        return super_staged

    def unstage(self) -> list[object] | DeviceStatus | StatusBase:  # type: ignore
        """Unstage the device."""
        super_unstage = super().unstage()
        status = self.on_unstage()  # pylint: disable=assignment-from-no-return
        if isinstance(status, StatusBase):
            return status
        return super_unstage

    def pre_scan(self) -> DeviceStatus | StatusBase | None:
        """Pre-scan function."""
        status = self.on_pre_scan()  # pylint: disable=assignment-from-no-return
        return status

    def trigger(self) -> DeviceStatus | StatusBase:
        """Trigger the device."""
        super_trigger = super().trigger()
        status = self.on_trigger()  # pylint: disable=assignment-from-no-return
        return status if status else super_trigger

    def complete(self) -> DeviceStatus | StatusBase:
        """Complete the device."""
        status = self.on_complete()  # pylint: disable=assignment-from-no-return
        if isinstance(status, StatusBase):
            return status
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def kickoff(self) -> DeviceStatus | StatusBase:
        """Kickoff the device."""
        status = self.on_kickoff()  # pylint: disable=assignment-from-no-return
        if isinstance(status, StatusBase):
            return status
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def stop(self, *, success: bool = False) -> None:
        """Stop the device.

        Args:
            success (bool): True if the action was successful, False otherwise.
        """
        self.on_stop()
        self.stopped = True  # Set stopped flag to True, in case a custom stop method listens to stopped property
        super().stop(success=success)

    ########################################
    # Utility Method to wait for signals   #
    ########################################

    def wait_for_condition(
        self,
        condition: Callable[[], bool],
        timeout: float,
        check_stopped: bool = False,
        interval: float = 0.05,
    ) -> bool:
        """
        Utility method to easily wait for signals or methods to reach an expected state.

        Args:
            condition (Callable): function that returns True if the condition is met, False otherwise
            timeout (float): timeout in seconds
            check_stopped (bool): True if stopped flag should be checked
            interval (float): interval in seconds

        Returns:
            bool: True if all signals are in the desired state, False if timeout is reached

        Example:
            >>> self.wait_for_condition(condition=my_condition, timeout=5, interval=0.05, check_stopped=True)
        """

        start_time = time.time()
        while time.time() < start_time + timeout:
            if condition() is True:
                return True
            if check_stopped is True and self.stopped is True:
                raise DeviceStoppedError(f"Device {self.name} has been stopped")
            time.sleep(interval)
        return False

    ########################################
    #  Beamline Specific Implementations   #
    ########################################

    def on_init(self) -> None:
        """
        Called when the device is initialized.

        No signals are connected at this point. If you like to
        set default values on signals, please use on_connected instead.
        """

    def on_connected(self) -> None:
        """
        Called after the device is connected and its signals are connected.
        Default values for signals should be set here.
        """

    def on_stage(self) -> DeviceStatus | StatusBase | None:
        """
        Called while staging the device.

        Information about the upcoming scan can be accessed from the scan_info (self.scan_info.msg) object.
        """

    def on_unstage(self) -> DeviceStatus | StatusBase | None:
        """Called while unstaging the device."""

    def on_pre_scan(self) -> DeviceStatus | StatusBase | None:
        """Called right before the scan starts on all devices automatically."""

    def on_trigger(self) -> DeviceStatus | StatusBase | None:
        """Called when the device is triggered."""

    def on_complete(self) -> DeviceStatus | StatusBase | None:
        """Called to inquire if a device has completed a scans."""

    def on_kickoff(self) -> DeviceStatus | StatusBase | None:
        """Called to kickoff a device for a fly scan. Has to be called explicitly."""

    def on_stop(self) -> None:
        """Called when the device is stopped."""



"""
Base classes for XIA xMAP and FalconX dxp system.
Falcon interfaces with the dxpSITORO epics driver, https://github.com/epics-modules/dxpSITORO.
xMAP interfaces with the dxp epics driver, https://github.com/epics-modules/dxp.

An example usage for a 4-element FalconX system. ::

    from ophyd import Component as Cpt
    from ophyd_devices.devices.dxp import Falcon, EpicsMCARecord, EpicsDXPFalcon
    from ophyd_devices.devices.areadetector.plugins import HDF5Plugin_V35 as HDF5Plugin

    class FalconX4(Falcon):

        # Attributes needed to be set for the ADBase class
        # Otherwise, ADBase will overwrite read_attrs and configuration_attrs
        # and as a result a .read() will return an empty dictionary
        # This is a shortcoming of the ADBase class and has to be fixed
        # on the highest level of the class hierarchy, one may also include specific signals of the Falcon
        # in the read_attrs and configuration_attrs, i.e. elapsed_real_time, elapsed_live_time, etc.
        _default_read_attrs = ("dxp1", "dxp2", "dxp3", "dxp4", "mca1", "mca2", "mca3", "mca4", "hdf")
        _default_configuration_attrs = ("dxp1", "dxp2", "dxp3", "dxp4", "mca1", "mca2", "mca3", "mca4", "hdf")

        # DXP parameters
        dxp1 = Cpt(EpicsDXPFalcon, "dxp1:")
        dxp2 = Cpt(EpicsDXPFalcon, "dxp2:")
        dxp3 = Cpt(EpicsDXPFalcon, "dxp3:")
        dxp4 = Cpt(EpicsDXPFalcon, "dxp4:")

        # MCA record with spectrum data
        mca1 = Cpt(EpicsMCARecord, "mca1")
        mca2 = Cpt(EpicsMCARecord, "mca2")
        mca3 = Cpt(EpicsMCARecord, "mca3")
        mca4 = Cpt(EpicsMCARecord, "mca4")

        # optionally with a HDF5 writer plugin
        hdf = Cpt(HDF5Plugin, "HDF1:")

    falcon = FalconX4("X07MB-SITORO:", name="falcon")
    falcon.collect_mode.put(0) # 0: MCA spectra, 1: MCA mapping
    falcon.preset_mode.put("Real time")
    falcon.preset_real_time.put(1)
    status = falcon.erase_start.set(1)
    status.wait()
    falcon.mca1.spectrum.get()

"""

from collections import OrderedDict

from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, Kind
from ophyd.areadetector import ADBase
from ophyd.areadetector import ADComponent as ADCpt
from ophyd.areadetector import EpicsSignalWithRBV
from ophyd.device import DynamicDeviceComponent as DCpt
from ophyd.mca import ROI as _ROI
from ophyd.mca import EpicsDXP, EpicsDXPBaseSystem, EpicsDXPMapping
from ophyd.mca import EpicsDXPMultiElementSystem as _EpicsDXPMultiElementSystem
from ophyd.mca import EpicsMCARecord as _EpicsMCARecord

__all__ = ("EpicsMCARecord", "EpicsDXP", "EpicsDXPFalcon", "Falcon", "Mercury", "xMAP")

# pylint: disable=protected-access


class ROI(_ROI):
    """ROI for DXP system with proper Kind settings."""

    # normal Components
    count = Cpt(EpicsSignalRO, "", lazy=True, kind=Kind.normal)
    net_count = Cpt(EpicsSignalRO, "N", lazy=True, kind=Kind.normal)

    # Config components
    preset_count = Cpt(EpicsSignal, "P", lazy=True, kind=Kind.config)
    bkgnd_chans = Cpt(EpicsSignal, "BG", lazy=True, kind=Kind.config)
    label = Cpt(EpicsSignal, "NM", lazy=True, kind=Kind.config)
    is_preset = Cpt(EpicsSignal, "IP", lazy=True, kind=Kind.config)
    hi_chan = Cpt(EpicsSignal, "HI", lazy=True, kind=Kind.config)
    lo_chan = Cpt(EpicsSignal, "LO", lazy=True, kind=Kind.config)


def add_rois(range_, **kwargs):
    """Add ROIs to the EpicsMCARecord."""
    defn = OrderedDict()

    for roi in range_:
        if not (0 <= roi < 32):
            raise ValueError("roi must be in the set [0,31]")

        attr = f"roi{roi}"
        defn[attr] = (ROI, f".R{roi}", kwargs)

    return defn


# predefined nunmber of ROIs per EpicsMCARecord
ROI_RANGE = range(0, 8)


class EpicsMCARecord(_EpicsMCARecord):
    """EpicsMCARecord with addtional fields"""

    # Calibration values
    calo = Cpt(EpicsSignal, ".CALO", kind=Kind.config)
    cals = Cpt(EpicsSignal, ".CALS", kind=Kind.config)
    calq = Cpt(EpicsSignal, ".CALQ", kind=Kind.config)
    tth = Cpt(EpicsSignal, ".TTH", kind=Kind.config)

    elapsed_real_time = Cpt(EpicsSignalRO, ".ERTM", kind=Kind.normal, auto_monitor=True)
    rois = DCpt(add_rois(ROI_RANGE, kind=Kind.normal), kind=Kind.normal)


class EpicsDXPFalcon(Device):
    """All high-level DXP parameters for each channel"""

    # Detection
    detection_filter = Cpt(EpicsSignalWithRBV, "DetectionFilter")
    detection_threshold = Cpt(EpicsSignalWithRBV, "DetectionThreshold")
    min_pulse_pair_separation = Cpt(EpicsSignalWithRBV, "MinPulsePairSeparation")

    # Pre-amp and energe range
    detector_polarity = Cpt(EpicsSignalWithRBV, "DetectorPolarity")
    decay_time = Cpt(EpicsSignalWithRBV, "DecayTime")
    risetime_optimization = Cpt(EpicsSignalWithRBV, "RisetimeOptimization")
    scale_factor = Cpt(EpicsSignalWithRBV, "ScaleFactor")

    # Presets
    preset_events = Cpt(EpicsSignalWithRBV, "PresetEvents")
    preset_mode = Cpt(EpicsSignalWithRBV, "PresetMode", string=True)
    preset_triggers = Cpt(EpicsSignalWithRBV, "PresetTriggers")
    preset_real_time = Cpt(EpicsSignalWithRBV, "PresetReal")

    # Couting statistics
    elapsed_live_time = Cpt(EpicsSignalRO, "ElapsedLiveTime", lazy=True)
    elapsed_real_time = Cpt(EpicsSignalRO, "ElapsedRealTime", lazy=True)
    elapsed_trigger_live = Cpt(EpicsSignalRO, "ElapsedTriggerLiveTime", lazy=True)
    triggers = Cpt(EpicsSignalRO, "Triggers", lazy=True)
    events = Cpt(EpicsSignalRO, "Events", lazy=True)
    input_count_rate = Cpt(EpicsSignalRO, "InputCountRate", kind=Kind.normal, auto_monitor=True)
    output_count_rate = Cpt(EpicsSignalRO, "OutputCountRate", kind=Kind.normal, auto_monitor=True)

    # Mapping
    current_pixel = Cpt(EpicsSignal, "CurrentPixel")

    # Diagnostic trace
    trace_data = Cpt(EpicsSignal, "TraceData")


class EpicsDXPFalconMultiElementSystem(EpicsDXPBaseSystem):
    """System-wide parameters as defined in dxpMED.template"""

    # Preset control
    preset_events = Cpt(EpicsSignal, "PresetEvents")
    preset_real_time = Cpt(EpicsSignal, "PresetReal")
    preset_mode = Cpt(EpicsSignal, "PresetMode", string=True)
    preset_triggers = Cpt(EpicsSignal, "PresetTriggers")

    # Acquisition control
    erase_all = Cpt(EpicsSignal, "EraseAll")
    erase_start = Cpt(EpicsSignal, "EraseStart", put_complete=True, trigger_value=1)
    start_all = Cpt(EpicsSignal, "StartAll", put_complete=True, trigger_value=1)
    stop_all = Cpt(EpicsSignal, "StopAll")

    # Status
    set_acquire_busy = Cpt(EpicsSignal, "SetAcquireBusy")
    acquire_busy = Cpt(EpicsSignal, "AcquireBusy")
    status_all = Cpt(EpicsSignal, "StatusAll")
    status_all_once = Cpt(EpicsSignal, "StatusAllOnce")
    acquiring = Cpt(EpicsSignal, "Acquiring")

    # Reading
    read_all = Cpt(EpicsSignal, "ReadAll", kind=Kind.omitted)
    read_all_once = Cpt(EpicsSignal, "ReadAllOnce", kind=Kind.omitted)

    # As a debugging note, if snl_connected is not '1', your IOC is
    # misconfigured:
    snl_connected = Cpt(EpicsSignal, "SNL_Connected")

    # High-level parameters
    copy_decay_time = Cpt(EpicsSignal, "CopyDecayTime", kind=Kind.omitted)
    copy_detection_filter = Cpt(EpicsSignal, "CopyDetectionFilter", kind=Kind.omitted)
    copy_detection_threshold = Cpt(EpicsSignal, "CopyDetectionThreshold", kind=Kind.omitted)
    copy_detector_polarity = Cpt(EpicsSignal, "CopyDetectorPolarity", kind=Kind.omitted)
    copy_min_pulse_pair_separation = Cpt(
        EpicsSignal, "CopyMinPulsePairSeparation", kind=Kind.omitted
    )
    copt_risetime_optimization = Cpt(EpicsSignal, "CopyRisetimeOptimization", kind=Kind.omitted)
    copy_scale_factor = Cpt(EpicsSignal, "CopyScaleFactor", kind=Kind.omitted)
    read_traces = Cpt(EpicsSignal, "ReadTraces", kind=Kind.omitted)

    # ROI and SCA
    copy_roi_channel = Cpt(EpicsSignal, "CopyROIChannel", kind=Kind.omitted)
    copy_roi_energy = Cpt(EpicsSignal, "CopyROIEnergy", kind=Kind.omitted)
    copy_roi_sca = Cpt(EpicsSignal, "CopyROI_SCA", kind=Kind.omitted)

    # do_* executes the process:
    do_read_all = Cpt(EpicsSignal, "DoReadAll", kind=Kind.omitted)
    do_status_all = Cpt(EpicsSignal, "DoStatusAll", kind=Kind.omitted)
    do_read_traces = Cpt(EpicsSignal, "DoReadTraces", kind=Kind.omitted)

    # Statistics
    dead_time = Cpt(EpicsSignal, "DeadTime")
    idead_time = Cpt(EpicsSignal, "IDeadTime")
    max_elapsed_live = Cpt(EpicsSignal, "MaxElapsedLive")
    max_elapsed_real = Cpt(EpicsSignal, "MaxElapsedReal")
    max_elapsed_trigger_live = Cpt(EpicsSignal, "MaxElapsedTriggerLive")
    max_triggers = Cpt(EpicsSignal, "MaxTriggers")
    max_events = Cpt(EpicsSignal, "MaxEvents")
    max_input_count_rate = Cpt(EpicsSignal, "MaxInputCountRate")
    max_output_count_rate = Cpt(EpicsSignal, "MaxOutputCountRate")


class EpicsDxpFalconMapping(EpicsDXPMapping):
    """Mapping mode parameters as defined in dxpMapping.template"""

    auto_apply = None
    apply = None
    nd_array_mode = Cpt(EpicsSignalWithRBV, "NDArrayMode")


class Falcon(EpicsDXPFalconMultiElementSystem, EpicsDxpFalconMapping, ADBase):
    """Falcon base device"""

    # attribute required by ADBase
    port_name = ADCpt(EpicsSignalRO, "Asyn.PORT", string=True)


class EpicsDXPMultiElementSystem(_EpicsDXPMultiElementSystem):
    """System-wide parameters as defined in dxpMED.template"""

    # Override some action signals, so calling `set`` method
    # returns a waitable Status object. Otherwise the Status object is immediately done.
    erase_start = Cpt(EpicsSignal, "EraseStart", put_complete=True, trigger_value=1)
    start_all = Cpt(EpicsSignal, "StartAll", put_complete=True, trigger_value=1)

    # mca.EpicsDXPMultiElementSystem maps the EPICS records under wrong names, i.e.
    # copy_adcp_ercent_rule, copy_roic_hannel and copy_roie_nergy
    copy_adc_percent_rule = Cpt(EpicsSignal, "CopyADCPercentRule")
    copy_roi_channel = Cpt(EpicsSignal, "CopyROIChannel")
    copy_roi_energy = Cpt(EpicsSignal, "CopyROIEnergy")


class Mercury(EpicsDXPMultiElementSystem, ADBase):
    """Mercury base device"""

    # attribute required by ADBase
    port_name = ADCpt(EpicsSignalRO, "Asyn.PORT", string=True)


class xMAP(EpicsDXPMultiElementSystem, EpicsDXPMapping, ADBase):
    """xMAP base device"""

    # attribute required by ADBase
    port_name = ADCpt(EpicsSignalRO, "Asyn.PORT", string=True)



# pylint: skip-file
import time
from unittest import mock

import pytest
from ophyd import DeviceStatus, Staged
from ophyd.utils.errors import RedundantStaging

from ophyd_devices.interfaces.base_classes.psi_device_base import PSIDeviceBase
from ophyd_devices.utils.errors import DeviceStopError, DeviceTimeoutError


@pytest.fixture
def detector_base():
    yield PSIDeviceBase(name="test_detector")


def test_detector_base_init(detector_base):
    assert detector_base.stopped is False
    assert detector_base.name == "test_detector"
    assert detector_base.staged == Staged.no
    assert detector_base.destroyed == False


def test_stage(detector_base):
    assert detector_base._staged == Staged.no
    assert detector_base.stopped is False
    detector_base._staged = Staged.no
    with mock.patch.object(detector_base, "on_stage") as mock_on_stage:
        rtr = detector_base.stage()
        assert isinstance(rtr, list)
        assert mock_on_stage.called is True
        with pytest.raises(RedundantStaging):
            detector_base.stage()
        detector_base._staged = Staged.no
        detector_base.stopped = True
        detector_base.stage()
        assert detector_base.stopped is False
        assert mock_on_stage.call_count == 2


# def test_stage(detector_base):
#     detector_base._staged = Staged.yes
#     with pytest.raises(RedundantStaging):
#         detector_base.stage()
#     assert detector_base.stopped is False
#     detector_base._staged = Staged.no
#     with (
#         mock.patch.object(detector_base.custom_prepare, "on_stage") as mock_on_stage,
#         mock.patch.object(detector_base.scaninfo, "load_scan_metadata") as mock_load_metadata,
#     ):
#         rtr = detector_base.stage()
#         assert isinstance(rtr, list)
#         mock_on_stage.assert_called_once()
#         mock_load_metadata.assert_called_once()
#         assert detector_base.stopped is False


# def test_pre_scan(detector_base):
#     with mock.patch.object(detector_base.custom_prepare, "on_pre_scan") as mock_on_pre_scan:
#         detector_base.pre_scan()
#         mock_on_pre_scan.assert_called_once()


# def test_trigger(detector_base):
#     status = DeviceStatus(detector_base)
#     with mock.patch.object(
#         detector_base.custom_prepare, "on_trigger", side_effect=[None, status]
#     ) as mock_on_trigger:
#         st = detector_base.trigger()
#         assert isinstance(st, DeviceStatus)
#         time.sleep(0.1)
#         assert st.done is True
#         st = detector_base.trigger()
#         assert st.done is False
#         assert id(st) == id(status)


# def test_unstage(detector_base):
#     detector_base.stopped = True
#     with (
#         mock.patch.object(detector_base.custom_prepare, "on_unstage") as mock_on_unstage,
#         mock.patch.object(detector_base, "check_scan_id") as mock_check_scan_id,
#     ):
#         rtr = detector_base.unstage()
#         assert isinstance(rtr, list)
#         assert mock_check_scan_id.call_count == 1
#         assert mock_on_unstage.call_count == 1
#         detector_base.stopped = False
#         rtr = detector_base.unstage()
#         assert isinstance(rtr, list)
#         assert mock_check_scan_id.call_count == 2
#         assert mock_on_unstage.call_count == 2


# def test_complete(detector_base):
#     status = DeviceStatus(detector_base)
#     with mock.patch.object(
#         detector_base.custom_prepare, "on_complete", side_effect=[None, status]
#     ) as mock_on_complete:
#         st = detector_base.complete()
#         assert isinstance(st, DeviceStatus)
#         time.sleep(0.1)
#         assert st.done is True
#         st = detector_base.complete()
#         assert st.done is False
#         assert id(st) == id(status)


# def test_stop(detector_base):
#     with mock.patch.object(detector_base.custom_prepare, "on_stop") as mock_on_stop:
#         detector_base.stop()
#         mock_on_stop.assert_called_once()
#         assert detector_base.stopped is True


# def test_check_scan_id(detector_base):
#     detector_base.scaninfo.scan_id = "abcde"
#     detector_base.stopped = False
#     detector_base.check_scan_id()
#     assert detector_base.stopped is True
#     detector_base.stopped = False
#     detector_base.check_scan_id()
#     assert detector_base.stopped is False


# def test_wait_for_signal(detector_base):
#     my_value = False

#     def my_callback():
#         return my_value

#     detector_base
#     status = detector_base.custom_prepare.wait_with_status(
#         [(my_callback, True)],
#         check_stopped=True,
#         timeout=5,
#         interval=0.01,
#         exception_on_timeout=None,
#     )
#     time.sleep(0.1)
#     assert status.done is False
#     # Check first that it is stopped when detector_base.stop() is called
#     detector_base.stop()
#     # some delay to allow the stop to take effect
#     time.sleep(0.15)
#     assert status.done is True
#     assert status.exception().args == DeviceStopError(f"{detector_base.name} was stopped").args
#     detector_base.stopped = False
#     status = detector_base.custom_prepare.wait_with_status(
#         [(my_callback, True)],
#         check_stopped=True,
#         timeout=5,
#         interval=0.01,
#         exception_on_timeout=None,
#     )
#     # Check that thread resolves when expected value is set
#     my_value = True
#     # some delay to allow the stop to take effect
#     time.sleep(0.15)
#     assert status.done is True
#     assert status.success is True
#     assert status.exception() is None

#     detector_base.stopped = False
#     # Check that wait for status runs into timeout with expectd exception
#     my_value = "random_value"
#     exception = TimeoutError("Timeout")
#     status = detector_base.custom_prepare.wait_with_status(
#         [(my_callback, True)],
#         check_stopped=True,
#         timeout=0.01,
#         interval=0.01,
#         exception_on_timeout=exception,
#     )
#     time.sleep(0.2)
#     assert status.done is True
#     assert id(status.exception()) == id(exception)
#     assert status.success is False


# def test_wait_for_signal_returns_exception(detector_base):
#     my_value = False

#     def my_callback():
#         return my_value

#     # Check that wait for status runs into timeout with expectd exception

#     exception = TimeoutError("Timeout")
#     status = detector_base.custom_prepare.wait_with_status(
#         [(my_callback, True)],
#         check_stopped=True,
#         timeout=0.01,
#         interval=0.01,
#         exception_on_timeout=exception,
#     )
#     time.sleep(0.2)
#     assert status.done is True
#     assert id(status.exception()) == id(exception)
#     assert status.success is False

#     detector_base.stopped = False
#     # Check that standard exception is thrown
#     status = detector_base.custom_prepare.wait_with_status(
#         [(my_callback, True)],
#         check_stopped=True,
#         timeout=0.01,
#         interval=0.01,
#         exception_on_timeout=None,
#     )
#     time.sleep(0.2)
#     assert status.done is True
#     assert (
#         status.exception().args
#         == DeviceTimeoutError(
#             f"Timeout error for {detector_base.name} while waiting for signals {[(my_callback, True)]}"
#         ).args
#     )
#     assert status.success is False



from unittest import mock

from ophyd_devices.utils.controller import Controller


def test_controller_off():
    controller = Controller(socket_cls=mock.MagicMock(), socket_host="dummy", socket_port=123)
    controller.on()
    with mock.patch.object(controller.sock, "close") as mock_close:
        controller.off()
        assert controller.sock is None
        assert controller.connected is False
        mock_close.assert_called_once()

        # make sure it is indempotent
        controller.off()
    controller._reset_controller()


def test_controller_on():
    socket_cls = mock.MagicMock()
    Controller._controller_instances = {}
    controller = Controller(socket_cls=socket_cls, socket_host="dummy", socket_port=123)
    controller.on()
    assert controller.sock is not None
    assert controller.connected is True
    socket_cls().open.assert_called_once()

    # make sure it is indempotent
    controller.on()
    socket_cls().open.assert_called_once()
    controller._reset_controller()
